/*MA****************************************************************/
/*                                                                 */
/*  FILE     draw-functions.c                                      */
/*  MODULE   draw-functions                                        */
/*  PROGRAM  vpf                                                   */
/*  AUTHOR   Helmut Schmid, IMS, University of Stuttgart           */
/*  PURPOSE  layout and draw routines                              */
/*                                                                 */
/*ME****************************************************************/

#include <tcl.h>
#include <tk.h>
#include <string.h>
#include <assert.h>

#include "basic-functions.h"
#include "parse-forest.h"
#include "draw-buffer.h"
#include "tag.h"
#include "draw-functions.h"

extern Tcl_Interp *Interp;  /* to be defined externally */


static void compute_size( NODE *node );
static void compute_pos( NODE *node );
static void draw_node( NODE *node );
static NODE *get_node_by_pos( NODE *node, long x, long y );

static const XY Margin        = {10, 10};
static const XY Square_Bracket= { 5,  2};
static const XY Curly_Bracket = {10,  2};
static const XY Feature_Sep   = { 5,  0};

static const int Tree_SepY    = 30;
static const int Dot_SepY     = 5;
static const int Tag_SepX     = 5;

static NODE *Root;
static NODE *Highlighted_Node = NULL;
static int Aligned=1;

static Tk_Font Canvas_Font;
static int Font_Height;
static int Digit_Height;
static char Display_Name[]=".d";


/*FA****************************************************************/
/*                                                                 */
/*  parse_forest_count                                             */
/*                                                                 */
/*FE****************************************************************/

int parse_forest_count( )

{
  return AllParses->value.data.node.number_of_daughters;
}


/*FA****************************************************************/
/*                                                                 */
/*  label_size                                                     */
/*                                                                 */
/*FE****************************************************************/

static XY label_size( char *string )

{
  XY size;

  size.x = Tk_TextWidth(Canvas_Font, string, strlen(string));
  size.y = Font_Height;

  return size;
}


/*FA****************************************************************/
/*                                                                 */
/*  tag_size                                                       */
/*                                                                 */
/*FE****************************************************************/

static XY tag_size( NODE *node )

{
  XY size;
  char *label=itoa(node->value.data.shared.tag_number);

  size.x = Tk_TextWidth(Canvas_Font, label, strlen(label));
  size.y = Digit_Height;

  return size;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_term_size                                              */
/*                                                                 */
/*FE****************************************************************/

static void compute_term_size( NODE *node )

{
  node->size = label_size(node->value.data.tnode.name);
  node->value.data.tnode.label_width = node->size.x;

  if (node->value.data.tnode.subnode != NULL) {
    compute_size(node->value.data.tnode.subnode);
    node->size.x += node->value.data.tnode.subnode->size.x + Feature_Sep.x;
    if (node->size.y < node->value.data.tnode.subnode->size.y)
      node->size.y = node->value.data.tnode.subnode->size.y;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_imploded_size                                          */
/*                                                                 */
/*FE****************************************************************/

static void compute_imploded_size( NODE *node )

{
  node->size.x = node->size.y = Digit_Height+1;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_row_size                                               */
/*                                                                 */
/*FE****************************************************************/

static void compute_row_size( NODE *node )

{
  int i,n;

  node->size.x = node->size.y = 0;
  for( i=n=0; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    compute_size(dnode);
    if (!(dnode->flags & HIDDEN_MASK)) {
      node->size.x += dnode->size.x;
      if (node->size.y < dnode->size.y)
	node->size.y = dnode->size.y;
      n++;
    }
  }
  if (n>1)
    node->size.x += (n-1) * Margin.x;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_disjunct_size                                          */
/*                                                                 */
/*FE****************************************************************/

static void compute_disjunct_size( NODE *node )

{
  compute_row_size(node);
  if (node->size.y == 0) {
    node->size.x = 2*Curly_Bracket.x + Margin.x/2;
    node->size.y = 2*Curly_Bracket.y + Margin.y;
    return;
  }

  node->size.x += 2 * Curly_Bracket.x;
  node->size.y += 2 * Curly_Bracket.y;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_column_size                                            */
/*                                                                 */
/*FE****************************************************************/

static void compute_column_size( NODE *node )

{
  int i, n;

  node->size.x = node->size.y = 0;

  for( i=n=0; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    compute_size(dnode);
    if (!(dnode->flags & HIDDEN_MASK)) {
      if (node->size.x < dnode->size.x)
	node->size.x = dnode->size.x;
      node->size.y += dnode->size.y;
      n++;
    }
  }

  if (n>1)
    node->size.y += (n-1) * Feature_Sep.y;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_fs_size                                                */
/*                                                                 */
/*FE****************************************************************/

static void compute_fs_size( NODE *node )

{
  int i, l1, l2;

  /* Compute unaligned size */
  compute_column_size( node );
  if (node->size.y == 0) {
    node->size.x = 2*Square_Bracket.x + Margin.x/2;
    node->size.y = 2*Square_Bracket.y + Margin.y;
    return;
  }

  node->size.x += 2 * Square_Bracket.x;
  node->size.y += 2 * Square_Bracket.y;

  if (Aligned) {
    /* Compute aligned size */
    for( l1=l2=i=0; i<node->value.data.node.number_of_daughters; i++) {
      NODE *dnode=node->value.data.node.daughter[i];
      if (dnode->value.type == terminal && 
	  dnode->value.data.tnode.subnode != NULL) {
	TNODE *tnode = &dnode->value.data.tnode;
	if (tnode->label_width > l1)
	  l1 = tnode->label_width;
	if (tnode->subnode->size.x > l2)
	  l2 = tnode->subnode->size.x;
      }
    }

    l2 = l1 + l2 + Feature_Sep.x + 2 * Square_Bracket.x;
    if (l2 > node->size.x)
      node->size.x = l2;

    /* Add padding space for alignment */
    for( i=0; i<node->value.data.node.number_of_daughters; i++) {
      NODE *dnode=node->value.data.node.daughter[i];
      if (dnode->value.type == terminal) {
	TNODE *tnode = &dnode->value.data.tnode;
	tnode->padding = l1 - tnode->label_width;
	dnode->size.x = node->size.x;
      }
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_tree_size                                              */
/*                                                                 */
/*FE****************************************************************/

static void compute_tree_size( NODE *node )

{
  int i,n;
  XY size;

  size.x = size.y = 0;
  compute_size(node->value.data.node.daughter[0]);
  node->size = node->value.data.node.daughter[0]->size;
  for( i=1, n=0; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    compute_size(dnode);
    if (!(dnode->flags & HIDDEN_MASK)) {
      size.x += dnode->size.x;
      if (size.y < dnode->size.y)
	size.y = dnode->size.y;
      n++;
    }
  }
  size.x += (n-1) * Margin.x;
  if (node->size.x < size.x)
    node->size.x = size.x;
  node->size.y += size.y + Tree_SepY;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_shared_size                                            */
/*                                                                 */
/*FE****************************************************************/

static void compute_shared_size( NODE *node )

{
  XY size = tag_size(node);

  node->value.data.shared.tag_size = size;

  compute_size(node->value.data.shared.subnode);
  node->size.x = node->value.data.shared.subnode->size.x + 
    node->value.data.shared.tag_size.x + Tag_SepX;
  node->size.y = node->value.data.shared.subnode->size.y;
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_size                                                   */
/*                                                                 */
/*FE****************************************************************/

static void compute_size( NODE *node )

{
  if (node->flags & HIDDEN_MASK)
    return;
  if (node->flags & TAG_MASK)
    node->size = tag_size(node);
  else {
    if (node->flags & IMPLODED_MASK)
      compute_imploded_size(node);
    else
      switch (node->value.type) {
      case terminal: compute_term_size(node); break;
      case tree:     compute_tree_size(node); break;
      case disjunct: compute_disjunct_size(node); break;
      case fs:       compute_fs_size(node); break;
      case row:      compute_row_size(node); break;
      case column:   compute_column_size(node); 
	node->size.x += 2*Margin.x;
	node->size.y += 2*Margin.y;
	break;
      case shared:   compute_shared_size(node); break;
      }
  }
}



/*FA****************************************************************/
/*                                                                 */
/*  compute_term_pos                                               */
/*                                                                 */
/*FE****************************************************************/

static void compute_term_pos( NODE *node )

{
  TNODE *tnode = &node->value.data.tnode;

  if (tnode->subnode != NULL) {
    tnode->subnode->pos.y = node->pos.y;
    tnode->subnode->pos.x = node->pos.x + tnode->label_width + 
      tnode->padding + Feature_Sep.x;
    compute_pos(tnode->subnode);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_horiz_pos                                              */
/*                                                                 */
/*FE****************************************************************/

static void compute_horiz_pos( NODE *node, XY pos )

{
  int i;

  for( i=0; i<node->value.data.node.number_of_daughters; i++)
    if (!(node->value.data.node.daughter[i]->flags & HIDDEN_MASK)) {
      NODE *dnode=node->value.data.node.daughter[i];
      dnode->pos = pos;
      compute_pos(dnode);
      pos.x += dnode->size.x + Margin.x;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_row_pos                                                */
/*                                                                 */
/*FE****************************************************************/

static void compute_row_pos( NODE *node )

{
  compute_horiz_pos( node, node->pos );
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_disjunct_pos                                           */
/*                                                                 */
/*FE****************************************************************/

static void compute_disjunct_pos( NODE *node )

{
  XY pos;

  pos = node->pos;
  pos.x += Curly_Bracket.x;
  pos.y += Curly_Bracket.y;
  compute_horiz_pos( node, pos );
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_tree_pos                                               */
/*                                                                 */
/*FE****************************************************************/

static void compute_tree_pos( NODE *node )

{
  NODE *pnode;
  XY pos;
  int i;
  long b;

  /* Center the parent node in its parcel */
  pnode = node->value.data.node.daughter[0];
  pnode->pos = node->pos;
  pnode->pos.x += (node->size.x - pnode->size.x) / 2;
  compute_pos(node->value.data.node.daughter[0]);

  b = (node->value.data.node.number_of_daughters-2) * Margin.x;
  for( i=1; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    if (!(dnode->flags & HIDDEN_MASK))
      b += dnode->size.x;
  }

  pos.x = node->pos.x + (node->size.x - b) / 2; /* center the daughter nodes */
  pos.y = node->pos.y + pnode->size.y + Tree_SepY;

  for( i=1; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    dnode->pos = pos;
    compute_pos(dnode);
    if (!(dnode->flags & HIDDEN_MASK))
      pos.x += node->value.data.node.daughter[i]->size.x + Margin.x;
  }
}



/*FA****************************************************************/
/*                                                                 */
/*  compute_vertic_pos                                             */
/*                                                                 */
/*FE****************************************************************/

static void compute_vertic_pos( NODE *node, XY pos )

{
  int i;

  for( i=0; i<node->value.data.node.number_of_daughters; i++)
    if (!(node->value.data.node.daughter[i]->flags & HIDDEN_MASK)) {
      NODE *dnode=node->value.data.node.daughter[i];
      dnode->pos = pos;
      compute_pos(dnode);
      pos.y += dnode->size.y + Feature_Sep.y;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_column_pos                                             */
/*                                                                 */
/*FE****************************************************************/

static void compute_column_pos( NODE *node )

{
  XY pos;

  pos = node->pos;
  pos.x += Margin.x;
  pos.y += Margin.y;
  compute_vertic_pos( node, pos );  
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_fs_pos                                                 */
/*                                                                 */
/*FE****************************************************************/

static void compute_fs_pos( NODE *node )

{
  XY pos;

  pos = node->pos;
  pos.x += Square_Bracket.x;
  pos.y += Square_Bracket.y;
  compute_vertic_pos( node, pos );  
}


/*FA****************************************************************/
/*                                                                 */
/*  compute_shared_pos                                             */
/*                                                                 */
/*FE****************************************************************/

static void compute_shared_pos( NODE *node )

{
  node->value.data.shared.subnode->pos = node->pos;
  compute_pos(node->value.data.shared.subnode);
}



/*FA****************************************************************/
/*                                                                 */
/*  compute_pos                                                    */
/*                                                                 */
/*FE****************************************************************/

static void compute_pos( NODE *node )

{
  if (node->flags & HIDDEN_MASK)
    return;
  if (node->flags & (TAG_MASK | IMPLODED_MASK))
    return;

  switch (node->value.type) {
  case terminal: compute_term_pos(node); break;
  case tree:     compute_tree_pos(node); break;
  case disjunct: compute_disjunct_pos(node); break;
  case fs:       compute_fs_pos(node); break;
  case row:      compute_row_pos(node); break;
  case column:   compute_column_pos(node); break;
  case shared:   compute_shared_pos(node); break;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  layout                                                         */
/*                                                                 */
/*FE****************************************************************/

static void layout()

{
  compute_size(Root);
  Root->pos = Margin;
  compute_pos(Root);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_edge                                                      */
/*                                                                 */
/*FE****************************************************************/

static void draw_edge( XY pos, XY pos2 )

{
  int d = pos.x - pos2.x;

  if (d < 3 && d > 0) {
    pos.x -= d/2;
    pos2.x = pos.x;
  }
  else if (d > -3 && d < 0) {
    pos.x += d/2;
    pos2.x = pos.x;
  }

  /* Sample draw command */
  /* e 188 297 212 337 */
  append_to_buffer("e ");
  append_to_buffer(itoa(pos.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(pos.y));
  append_to_buffer(" ");
  append_to_buffer(itoa(pos2.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(pos2.y));
  append_to_buffer("\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_bracket                                                   */
/*                                                                 */
/*FE****************************************************************/

static void draw_bracket( char *command, XY pos, int length )

{
  /* Sample draw command */
  /* lcb 188 297 45 */
  append_to_buffer(command);
  append_to_buffer(" ");
  append_to_buffer(itoa(pos.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(pos.y));
  append_to_buffer(" ");
  append_to_buffer(itoa(length));
  append_to_buffer("\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_highlighting                                              */
/*                                                                 */
/*FE****************************************************************/

static void draw_highlighting( NODE *node )

{
  /* Sample draw command */
  /* h 88 97 253 198 */
  append_to_buffer("h ");
  append_to_buffer(itoa(node->pos.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.y));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.x + node->size.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.y + node->size.y));
  append_to_buffer("\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_label                                                     */
/*                                                                 */
/*FE****************************************************************/

static void draw_label( char *string, XY pos )

{
  /* Sample draw command */
  /* l 88 97 {Number} */
  append_to_buffer("l ");
  append_to_buffer(itoa(pos.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(pos.y));
  append_to_buffer(" {");
  append_to_buffer(string);
  append_to_buffer("}\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  gen_draw_tag                                                   */
/*                                                                 */
/*FE****************************************************************/

static void gen_draw_tag( char *command, int x1, int y1, int x2, int y2, int n)

{
  /* Sample draw command */
  /* t 253 461 5 */
  append_to_buffer(command);
  append_to_buffer(" ");
  append_to_buffer(itoa(x1));
  append_to_buffer(" ");
  append_to_buffer(itoa(y1));
  append_to_buffer(" ");
  append_to_buffer(itoa(x2));
  append_to_buffer(" ");
  append_to_buffer(itoa(y2));
  append_to_buffer(" ");
  append_to_buffer(itoa(n));
  append_to_buffer("\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_tag                                                       */
/*                                                                 */
/*FE****************************************************************/

static void draw_tag( NODE *node )

{
  gen_draw_tag("t", node->pos.x, node->pos.y, node->pos.x+node->size.x, 
	       node->pos.y+node->size.y, node->value.data.shared.tag_number);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_imploded                                                  */
/*                                                                 */
/*FE****************************************************************/

static void draw_imploded( NODE *node )

{
  /* Sample draw command */
  /* i 253 461 */
  append_to_buffer("i ");
  append_to_buffer(itoa(node->pos.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.y));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.x + node->size.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(node->pos.y + node->size.y));
  append_to_buffer("\n");
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_term                                                      */
/*                                                                 */
/*FE****************************************************************/

static void draw_term( NODE *node )

{
  TNODE *tnode = &node->value.data.tnode;

  draw_label( tnode->name, node->pos );

  if (tnode->subnode != NULL)
    draw_node(tnode->subnode);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_row                                                       */
/*                                                                 */
/*FE****************************************************************/

static void draw_row( NODE *node )

{
  int i;

  for( i=0; i<node->value.data.node.number_of_daughters; i++)
    if (!(node->value.data.node.daughter[i]->flags & HIDDEN_MASK))
      draw_node(node->value.data.node.daughter[i]);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_disjunct                                                  */
/*                                                                 */
/*FE****************************************************************/

static void draw_disjunct( NODE *node )

{
  XY pos;

  pos = node->pos;
  draw_bracket( "lcb", pos, node->size.y );
  pos.x += node->size.x;
  draw_bracket( "rcb", pos, node->size.y );
  draw_row( node );
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_column                                                    */
/*                                                                 */
/*FE****************************************************************/

static void draw_column( NODE *node )

{
  int i;

  for( i=0; i<node->value.data.node.number_of_daughters; i++)
    if (!(node->value.data.node.daughter[i]->flags & HIDDEN_MASK))
      draw_node(node->value.data.node.daughter[i]);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_fs                                                        */
/*                                                                 */
/*FE****************************************************************/

static void draw_fs( NODE *node )

{
  XY pos;

  pos = node->pos;
  draw_bracket( "lsb", pos, node->size.y );
  pos.x += node->size.x;
  draw_bracket( "rsb", pos, node->size.y );
  draw_column( node );
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_tree                                                      */
/*                                                                 */
/*FE****************************************************************/

static void draw_tree( NODE *node )

{
  int i;
  XY ppos, dpos;
  NODE *pnode=node->value.data.node.daughter[0];

  ppos.x = node->pos.x + node->size.x / 2;
  ppos.y = node->pos.y + pnode->size.y + Dot_SepY;

  for( i=1; i<node->value.data.node.number_of_daughters; i++) {
    NODE *dnode=node->value.data.node.daughter[i];
    dpos.x = dnode->pos.x + dnode->size.x / 2;
    dpos.y = dnode->pos.y - Dot_SepY;
    draw_edge( ppos, dpos );
  }

  for( i=0; i<node->value.data.node.number_of_daughters; i++)
    if (!(node->value.data.node.daughter[i]->flags & HIDDEN_MASK))
      draw_node(node->value.data.node.daughter[i]);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_shared                                                    */
/*                                                                 */
/*FE****************************************************************/

static void draw_shared( NODE *node )

{
  gen_draw_tag("tt", 
	       node->pos.x + node->size.x - node->value.data.shared.tag_size.x,
	       node->pos.y, node->pos.x + node->size.x, 
	       node->pos.y + node->value.data.shared.tag_size.y,
	       node->value.data.shared.tag_number);
  draw_node(node->value.data.shared.subnode);
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_node                                                      */
/*                                                                 */
/*FE****************************************************************/

static void draw_node( NODE *node )

{
  if (node->flags & HIDDEN_MASK)
    return;
  if (node->flags & HIGHLIGHTED_MASK)
    draw_highlighting(node);
  if (node->flags & TAG_MASK)
    draw_tag(node);
  else {
    if (node->flags & IMPLODED_MASK)
      draw_imploded(node);
    else
      switch (node->value.type) {
      case terminal: draw_term(node); break;
      case tree:     draw_tree(node); break;
      case disjunct: draw_disjunct(node); break;
      case fs:       draw_fs(node); break;
      case row:      draw_row(node); break;
      case column:   draw_column(node); break;
      case shared:   draw_shared(node); break;
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  gen_draw_commands                                              */
/*                                                                 */
/*FE****************************************************************/

static char *gen_draw_commands()

{
  XY size;
  char *commands;

  clear_buffer();

  size = Root->size;
  size.x += 2 * Margin.x;
  size.y += 2 * Margin.y;

  /* display size command */
  /* d 545 698 */
  append_to_buffer("d ");
  append_to_buffer(itoa(size.x));
  append_to_buffer(" ");
  append_to_buffer(itoa(size.y));
  append_to_buffer("\n");
  
  draw_node( Root );
  commands = get_buffer();
  return commands;
}


/*FA****************************************************************/
/*                                                                 */
/*  set_font                                                       */
/*                                                                 */
/*FE****************************************************************/

static void set_font()

{
  Tk_Window mw,fw;
  Tk_Uid fontname;
  Tk_FontMetrics fm;

  if (!Interp) error("in function layout: no Tcl interpreter");
  mw = Tk_MainWindow(Interp);
  if (!mw) error("in function layout: Can't find main window of application");
  fw = Tk_NameToWindow(Interp, Display_Name, mw);
  if (!fw) error("in function layout: Can't find frame");
  fontname = Tk_GetOption(fw, "font", "Display");
  if (!fontname) error("in function layout: Can't determine font");
  Canvas_Font = Tk_GetFont(Interp, fw, fontname);
  if (!fontname) error("in function layout: Can't find font");
  Tk_GetFontMetrics(Canvas_Font, &fm);
  Font_Height = fm.linespace;
  Digit_Height = fm.ascent;
}


/*FA****************************************************************/
/*                                                                 */
/*  free_font                                                      */
/*                                                                 */
/*FE****************************************************************/

static void free_font()

{
  Tk_FreeFont( Canvas_Font );
}


/*FA****************************************************************/
/*                                                                 */
/*  draw_commands                                                  */
/*                                                                 */
/*FE****************************************************************/

char *draw_commands( int index )

{
  char *commands;

  assert(index>=0 && index < AllParses->value.data.node.number_of_daughters);
  set_font();
  Root = AllParses->value.data.node.daughter[index];
  layout();
  commands = gen_draw_commands();
  free_font();

  return commands;
}


/*FA****************************************************************/
/*                                                                 */
/*  get_node_by_pos_term                                           */
/*                                                                 */
/*FE****************************************************************/

static NODE *get_node_by_pos_term( NODE *node, long x, long y )

{
  if (node->value.data.tnode.subnode) {
    NODE *result = get_node_by_pos(node->value.data.tnode.subnode, x, y);
    if (result)
      return result;
    else
      return node->value.data.tnode.subnode;
  }
  return node;
}


/*FA****************************************************************/
/*                                                                 */
/*  get_node_by_pos_shared                                         */
/*                                                                 */
/*FE****************************************************************/

static NODE *get_node_by_pos_shared( NODE *node, long x, long y )

{
  NODE *result = get_node_by_pos(node->value.data.shared.subnode, x, y);
  if (result)
    return result;
  else
    return node;
}


/*FA****************************************************************/
/*                                                                 */
/*  get_node_by_pos_nterm                                          */
/*                                                                 */
/*FE****************************************************************/

static NODE *get_node_by_pos_nterm( NODE *node, long x, long y )

{
  int i;

  for( i=0; i<node->value.data.node.number_of_daughters; i++) {
    NODE *result = get_node_by_pos(node->value.data.node.daughter[i], x, y);
    if (result)
      return result;
  }
  return node;
}


/*FA****************************************************************/
/*                                                                 */
/*  get_node_by_pos                                                */
/*                                                                 */
/*FE****************************************************************/

static NODE *get_node_by_pos( NODE *node, long x, long y )

{
  if (x >= node->pos.x &&
      x <= node->pos.x + node->size.x &&
      y >= node->pos.y &&
      y <= node->pos.y + node->size.y) {
    if (node->flags & (TAG_MASK | IMPLODED_MASK))
      return node;
    else if (node->value.type == terminal)
      return get_node_by_pos_term(node, x, y);
    else if (node->value.type == shared)
      return get_node_by_pos_shared(node, x, y);
    else
      return get_node_by_pos_nterm(node, x, y);
  }
  else
    return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  highlight_node                                                 */
/*                                                                 */
/*FE****************************************************************/

void highlight_node( NODE *node )

{
  if (Highlighted_Node)
    Highlighted_Node->flags ^= HIGHLIGHTED_MASK;  /* clear bit */
  Highlighted_Node = node;
  if (node)
    node->flags |= HIGHLIGHTED_MASK;              /* set bit */
}


/*FA****************************************************************/
/*                                                                 */
/*  set_highlighted                                                */
/*                                                                 */
/*FE****************************************************************/

int set_highlighted( long x, long y )

{
  NODE *node=get_node_by_pos( Root, x, y );

  if (Highlighted_Node == node)
    return 0;  /* nothing has changed */

  highlight_node( node );
  return 1;
}


/*FA****************************************************************/
/*                                                                 */
/*  toggle_imploded                                                */
/*                                                                 */
/*FE****************************************************************/

void toggle_imploded( long x, long y )

{
  NODE *node=get_node_by_pos( Root, x, y );

  if (node) {
    if (node->flags & TAG_MASK)
      set_selected(node);
    else
      node->flags ^= IMPLODED_MASK;            /* toggle bit */
  }
  highlight_node(node);
}
