
/*MA****************************************************************/
/*                                                                 */
/*  FILE      hashing.c                                            */
/*  AUTHOR    Helmut Schmid                                        */
/*  PURPOSE   general hashing module                               */
/*                                                                 */
/*ME****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "hashing.h"


#define HASH_CONSTANT  2654435769UL    /* (sqrt(5) - 1) / 2 */
#define START_SHIFT_FACTOR  (sizeof(unsigned long)*8 - 12) 
#define MIN_HASH_TABLE_SIZE 4096   /* 2^12 */
#define MAX_LOAD  4
/* rehashing if number of hashed entries  */
/* exceeds Hash_Table_Size * MAX_LOAD */

void rehash( HASH_TABLE *ht );



/*FA****************************************************************/
/*                                                                 */
/*  space_error                                                    */
/*                                                                 */
/*FE****************************************************************/

void space_error( char *function )

{
  fprintf(stderr,"\nERROR: not enough space in function %s\n",function);
  fprintf(stderr,"aborted!\n");
  exit(1);
}


/*FA****************************************************************/
/*                                                                 */
/*  make_hash_table                                                */
/*                                                                 */
/*FE****************************************************************/

HASH_TABLE *make_hash_table( hashf, cmpf )

KEY_TYPE (*hashf)( void *data );
int (*cmpf)( void *data1, void *data2 );

{
  long i;
  HASH_TABLE *ht;

  if ((ht = malloc(sizeof(HASH_TABLE))) == NULL)
    space_error("make_hash_table");
  
  ht->size = MIN_HASH_TABLE_SIZE;
  if ((ht->entry = malloc(sizeof(HASH_ELEM*) * ht->size)) == NULL)
    space_error("make_hash_table");
  
  ht->shift_factor = START_SHIFT_FACTOR;
  ht->number_of_hashes = 0;
  ht->hash_function = hashf;
  ht->identical_data = cmpf;
  for( i=0; i<ht->size; i++)
    ht->entry[i] = NULL;

  return ht;
}


/*FA****************************************************************/
/*                                                                 */
/*  free_hash_table                                                */
/*                                                                 */
/*FE****************************************************************/

void free_hash_table( HASH_TABLE *ht )

{
  int i;
  HASH_ELEM *elem, *p;

  for( i=0; i<ht->size; i++)
    for( elem=ht->entry[i]; elem; elem=p) {
      p = elem->next;
      free(elem);
    }
  free(ht->entry);
  free(ht);
}    

  

/*FA****************************************************************/
/*                                                                 */
/*  compute_key                                                    */
/*                                                                 */
/*FE****************************************************************/

inline KEY_TYPE compute_key( HASH_TABLE *ht, void *data )

{
  KEY_TYPE key;

  key = ht->hash_function( data );
  key = (key * HASH_CONSTANT) >> ht->shift_factor;

  return key;
}



/*FA****************************************************************/
/*                                                                 */
/*  lookup_data                                                    */
/*                                                                 */
/*FE****************************************************************/

void *lookup_data( HASH_TABLE *ht, void *data )

{
  HASH_ELEM *elem;
  KEY_TYPE  key;

  key = compute_key( ht, data );

  for( elem=ht->entry[key]; elem; elem=elem->next)
    if (ht->identical_data(elem->data, data))
      return(elem->data);

  return(NULL);
}



/*FA****************************************************************/
/*                                                                 */
/*  hash_data                                                      */
/*                                                                 */
/*FE****************************************************************/

int hash_data( HASH_TABLE *ht, void *data )

{
  HASH_ELEM *elem;
  KEY_TYPE  key;

  key = compute_key( ht, data );
  assert(key < ht->size);
  for( elem=ht->entry[key]; elem; elem=elem->next)
    if (ht->identical_data(elem->data, data))
      return 1;

  elem = malloc(sizeof(HASH_ELEM));
  elem->data = data;
  elem->next = ht->entry[key];
  ht->entry[key] = elem;

  if (++ht->number_of_hashes > ht->size * MAX_LOAD)
    rehash( ht );

  return 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  remove_data                                                    */
/*                                                                 */
/*FE****************************************************************/

int remove_data( HASH_TABLE *ht, void *data )

{
  HASH_ELEM **p, *elem;
  KEY_TYPE  key;

  key = compute_key( ht, data );

  for( p=&ht->entry[key]; *p; p=&(*p)->next)
    if (ht->identical_data((*p)->data, data)) {
      elem = *p;
      *p = (*p)->next;
      free(elem);
      if (--ht->number_of_hashes < ht->size / MAX_LOAD)
	rehash( ht );
      return(1);
    }

  return(0);
}


/*FA****************************************************************/
/*                                                                 */
/*  rehash                                                         */
/*                                                                 */
/*FE****************************************************************/

void rehash( HASH_TABLE *ht )

{
  long i, old_size;
  KEY_TYPE key;
  HASH_ELEM **old_table, *elem, *p;


  old_size = ht->size;
  old_table = ht->entry;

  if (ht->number_of_hashes > ht->size * MAX_LOAD) {
    if ((ht->shift_factor-=3) <= 10)
      {
	fprintf(stderr,"\nERROR: not enough memory\n");
	exit(1);
      }
    ht->size <<= 3;  /* 8 times as large */
  }
  else if (ht->number_of_hashes < ht->size / MAX_LOAD &&
	   ht->size > MIN_HASH_TABLE_SIZE) {
    ht->size >>= 3;
    if (ht->size < MIN_HASH_TABLE_SIZE)
      ht->size = MIN_HASH_TABLE_SIZE;
  }
  else
    return;


  /* Make a new, larger hash table. */

  if ((ht->entry = malloc (sizeof(HASH_ELEM*) * ht->size)) == NULL)
    space_error("rehash");
  for( i=0; i<ht->size; i++)
    ht->entry[i] = NULL;

  /* Rehash the XXXs from the old table into the new one. */

  for( i=0; i<old_size; i++)
    for( elem=old_table[i]; elem; elem=p)
      {
	p = elem->next;
	key = compute_key( ht, elem->data );
	elem->next = ht->entry[key];
	ht->entry[key] = elem;
      }
  
  free(old_table);
}

  

/*FA****************************************************************/
/*                                                                 */
/*  hash_table_stat                                                */
/*                                                                 */
/*FE****************************************************************/

void hash_table_stat( HASH_TABLE *ht )

{
  long i;
  HASH_ELEM *elem;
  int *count, n, max=0;;

  for( i=0; i<ht->size; i++)
    {
      for( n=0, elem=ht->entry[i]; elem; elem=elem->next )
	n++;
      if (max<n)
	max = n;
    }

  count = malloc(sizeof(int) * (max+1));
  for( n=0; n<=max; n++)
    count[n] = 0;

  for( i=0; i<ht->size; i++)
    {
      for( n=0, elem=ht->entry[i]; elem; elem=elem->next )
	n++;
      count[n]++;
    }

  fprintf(stderr,"\nXXX hash table statistics:\n\n");;
  for( n=0; n<=max; n++)
    if (count[n] > 0)
      fprintf(stderr,"%3d elements: %d times\n", n, count[n]);

  free(count);
}
