/*MA****************************************************************/
/*                                                                 */
/*  FILE     parse-forest.c                                        */
/*  MODULE   parse-forest                                          */
/*  PROGRAM  vpf                                                   */
/*  AUTHOR   Helmut Schmid, IMS, University of Stuttgart           */
/*  PURPOSE  routines for reading, storing and accessing the parse */
/*           forests                                               */
/*                                                                 */
/*ME****************************************************************/

#include <assert.h>

#include "basic-functions.h"
#include "parse-forest.h"
#include "tag.h"

NODE *AllParses;


/*FA****************************************************************/
/*                                                                 */
/*  parse_forest_init                                              */
/*                                                                 */
/*FE****************************************************************/

void parse_forest_init()

{
  tag_init();
}


/*FA****************************************************************/
/*                                                                 */
/*  append                                                         */
/*                                                                 */
/*FE****************************************************************/

NODE_LIST *append( NODE *node, NODE_LIST *nlist )

{
  NODE_LIST *result;

  GET_SPACE(result, NODE_LIST);
  result->node = node;
  result->next = nlist;
  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  alloc_node                                                     */
/*                                                                 */
/*FE****************************************************************/

static NODE *alloc_node( NODE_TYPE type )

{
  NODE *result;

  GET_SPACE(result, NODE);
  result->flags = 0;
  result->value.type = type;

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  new_node                                                       */
/*                                                                 */
/*FE****************************************************************/

NODE *new_node( NODE_LIST *nlist, NODE_TYPE type )

{
  NODE_LIST *p;
  NODE *result, **daughter;
  int n;

  /* count the daughter nodes */
  for( n=0,p=nlist; p; p=p->next,n++) ;

  /* allocate memory */
  result = alloc_node( type );
  result->value.data.node.number_of_daughters = n;
  GET_SPACE1(daughter, NODE*, n);
  result->value.data.node.daughter = daughter;
  result->flags = 0;

  /* copy the node references to the array */
  for( n=0; nlist; nlist=p, n++) {
    daughter[n] = nlist->node;
    p = nlist->next;
    free(nlist);  /* free the space allocated for the list */
  }

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  new_tnode                                                      */
/*                                                                 */
/*FE****************************************************************/

NODE *new_tnode( char *string, NODE *node)

{
  NODE *result;

  /* allocate memory */
  result = alloc_node( terminal );
  result->value.data.tnode.name = string;
  result->value.data.tnode.padding = 0;
  result->value.data.tnode.subnode = node;

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  new_shared_node                                                */
/*                                                                 */
/*FE****************************************************************/

NODE *new_shared_node( int tag, NODE *node )

{
  NODE *result=alloc_node( shared );

  result->value.data.shared.tag_number = tag;
  result->value.data.shared.subnode = node;
  if (node)
    insert_tagged_node(tag, node, result);
  else
    result->flags |= TAG_MASK;

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  print_daughters                                                */
/*                                                                 */
/*FE****************************************************************/
static void print_node( int selected, NODE *node, FILE *file );

static void print_daughters( int selected, NTNODE *node, char lb, char rb, 
			     FILE *file )

{
  int i;
  int flag=(selected && lb=='{');

  if (!flag)
    fputc(lb, file);
  for( i=0; i<node->number_of_daughters; i++) {
    if (flag) {
      if (!(node->daughter[i]->flags & (IMPLODED_MASK | TAG_MASK))) {
	/* print the first unimploded node and return */
	print_node(selected, node->daughter[i], file);
	return;
      }
      else
	continue; /* Don't print */
    }
    else
      print_node(selected, node->daughter[i], file);
  }
  if (!flag)
    fputc(rb, file);
}


/*FA****************************************************************/
/*                                                                 */
/*  print_tag                                                      */
/*                                                                 */
/*FE****************************************************************/

static void print_tag( int selected, NODE *node, FILE *file )

{
  fprintf(file,"#%d", node->value.data.shared.tag_number);

  if (!(node->flags & TAG_MASK)) {
    fputc('=',file);
    print_node(selected, node->value.data.shared.subnode, file);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  print_node                                                     */
/*                                                                 */
/*FE****************************************************************/

static void print_node( int selected, NODE *node, FILE *file )

{
  if (node->flags & IMPLODED_MASK)
    fputs(" #i",file);
  if (node->flags & HIDDEN_MASK)
    fputs(" #h",file);

  switch (node->value.type) {
  case terminal:
    fprintf(file,"%s", quote(node->value.data.tnode.name, (char)'"'));
    if (node->value.data.tnode.subnode == NULL)
      fprintf(file," ");
    else {
      fprintf(file,"=");
      print_node(0, node->value.data.tnode.subnode, file);
    }
    break;
  case tree:
    print_daughters(selected, &node->value.data.node, '(', ')', file);
    break;
  case disjunct:
    print_daughters(selected, &node->value.data.node, '{', '}', file);
    break;
  case fs:
    print_daughters(selected, &node->value.data.node, '[', ']', file);
    break;
  case column:
    print_daughters(selected, &node->value.data.node, '|', '|', file);
    break;
  case row:
    print_daughters(selected, &node->value.data.node, '-', '-', file);
    break;
  case shared:
    print_tag(selected, node, file);
    break;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  print_parses                                                   */
/*                                                                 */
/*FE****************************************************************/

void print_parses( FILE *file, int selected )

{
  int i;

  for( i=0; i<AllParses->value.data.node.number_of_daughters; i++) {
    NODE *node=AllParses->value.data.node.daughter[i];
    if (selected && (node->flags & IMPLODED_MASK))
      continue;
    print_node( selected, node, file );
    fputc('\n', file);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  node_references                                                */
/*                                                                 */
/*FE****************************************************************/

void node_references( NODE *node )

{
  if (node == NULL) return;

  if (node->flags & TAG_MASK) {
    if (node->value.data.shared.subnode == NULL) {
      int n = node->value.data.shared.tag_number;
      node->value.data.shared.subnode = get_node_by_tag(n);
      if (node->value.data.shared.subnode == NULL)
	error2("undefined tag", itoa(n));
    }
  }
  else if (node->value.type == terminal)
    node_references(node->value.data.tnode.subnode);
  else if (node->value.type == shared)
    node_references(node->value.data.shared.subnode);
  else {
    int i;
    for( i=0; i<node->value.data.node.number_of_daughters; i++)
      node_references(node->value.data.node.daughter[i]);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  insert_node_references                                         */
/*                                                                 */
/*FE****************************************************************/

void insert_node_references()

{
  node_references(AllParses);
}
